/*
 * Grundkurs Informatik - Beispielprogramm zur LZW-Datenkompression
 * H. Ernst, J. Schmidt 2016
 *
 * Ackermann-Funktion rekursiv und iterativ
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#define SMAX 1000000

// wenn FAST = 1: verwende geschlossene Lösungen in iterativer Variante
// wenn FAST = 0: berechne alles komplett
#define FAST 0

//***********************************************************************
// Rekursive Berechnung der Ackermann-Funktion
//-----------------------------------------------------------------------
int ak_r(int x, int y) {
	if (x == 0)
		return (y + 1); // Abbruchkriterium

	if (y == 0)
		return (ak_r(x - 1, 1)); // einfache Rekursion

	return (ak_r(x - 1, ak_r(x, y - 1))); // doppelte Rekursion
}

//***********************************************************************
// Iterative Berechnung der Ackermann-Funktion
// hier wurden auch die Spezialfälle A(1,y), A(2,y) und A(3,y) berücksichtigt,
// für die es geschlossene Lösungen gibt.
// die würde man natürlich auch bei der rekursiven Variante zur Beschleunigung
// anwenden.
// Möchte man das nicht, lässt sich diese Erweiterung über das define FAST abschalten
//-----------------------------------------------------------------------
int ak_i(int x, int y) {
	int s[SMAX + 2], sp = 0; // Stack und Stack-Pointer
	int i, zwischen; // Zähler, Zwischenergebnis
	s[sp++] = x;
	s[sp++] = y; // x und y in Stack eintragen

	while (sp != 1) {
		y = s[--sp];
		x = s[--sp]; // x und y aus Stack holen

		if (x == 0)
			s[sp++] = y + 1; // nur y+1 in Stack eintragen
#if FAST
		else if (x == 1) {   // Spezialfall A(1, y) = y + 2 lässt sich geschlossen angeben
			s[sp++] = y + 2;
		}
		else if (x == 2) {   // Spezialfall A(2, y) = 2y + 3 lässt sich geschlossen angeben
			s[sp++] = 2*y + 3;
		}
		else if (x == 3) {   // Spezialfall A(3, y) = 2^{y+3} - 3 lässt sich geschlossen angeben
			zwischen = 2;
			for(i = 1; i < y+3; i++) zwischen = zwischen * 2;
			s[sp++] = zwischen - 3;
		}
#endif
		else if (y == 0) {
			s[sp++] = x - 1;
			s[sp++] = 1;
		} // x-1 und 1 in Stack
		else {
			s[sp++] = x - 1;
			s[sp++] = x;
			s[sp++] = y - 1;
		} // x-1, x, y-1 in Stack

		if (sp >= SMAX)
			return -1; // Stack-Überlauf
	}

	return s[--sp]; // Ergebnis: letzter Stack-Eintrag
}

//***********************************************************************
// Hauptprogramm
//***********************************************************************
int main() {
	int x = 1, y = 1, a = 0; // Parameter deklarieren und vorbesetzen
	time_t t; // Zeit

	printf("\n\nACKERMANN-FUNKTION\n");
	while (x > 0 && y > 0) { // Solange x und y nicht 0 sind
		printf("\nx, y = ");
		scanf("%d %d", &x, &y); // Eingabe von x und y
		t = clock(); // Anfangszeit merken
		printf("\nak_r = %d ", ak_r(x, y)); // ak(x,y) rekursiv
		printf("(%5.2f sec)\n", (float) difftime(clock(), t) / CLOCKS_PER_SEC);

		t = clock(); // Anfangszeit merken
		a = ak_i(x, y); // ak(x,y) iterativ

		if (a < 0)
			printf("\nStack-Überlauf!!");
		else
			printf("\nak_i = %d ", a);

		printf("(%5.2f sec)\n", (float) difftime(clock(), t) / CLOCKS_PER_SEC);
	}

	return EXIT_SUCCESS;
}